__all__ = [
    "run_inference_for_video", 
    "load_graph", 
    "videoframe", 
    "draw_bbox", 
    "class_int_to_text", 
    "object_to_json_style", 
    "object_list"]

import sys
import os
import cv2
import re

from functools import lru_cache
import tensorflow as tf
import numpy as np
from operator import itemgetter

from . import gpu
from PIL import Image, ImageDraw, ImageFont
from google.protobuf import text_format

from ..protos import label_map_pb2


def gpu_selector(gpu_num=None):
    os.environ["CUDA_VISIBLE_DEVICES"] = str(gpu_num)
    if gpu_num == None:
        try:
            gpu_info = gpu.Info()
            if len(gpu_info.memory) > 1:
                gpu_info.memory = [i for i in gpu_info.memory if int(i.size) - int(i.used)> 6000]
                gpu_info.memory = sorted(gpu_info.memory, key=itemgetter(1))
            else:
                return "0"
        except:
            return None
        os.environ["CUDA_VISIBLE_DEVICES"] = str(gpu_info.memory[0].num)  # "0, 1" for multiple
        return gpu_info.memory[0].num
    return gpu_num

def run_inference_for_video(file_name, graph, gpu_num=None, allow_growth=True, process_gpu_memory=5000):
    config = tf.compat.v1.ConfigProto()
    gpu_num = gpu_selector(gpu_num)
    if gpu_num != None:
        if allow_growth:
            config.gpu_options.allow_growth = True
        else:
            gpu_info = gpu.Info()
            per_gpu_memory = int(process_gpu_memory) / int(gpu_info.memory[int(gpu_num)].size)
            config.gpu_options.per_process_gpu_memory_fraction = per_gpu_memory


    image_np_list = videoframe(file_name)

    with tf.Session(config=config, graph=graph) as sess:
        # Get handles to input and output tensors
        ops = tf.compat.v1.get_default_graph().get_operations()
        all_tensor_names = {output.name for op in ops for output in op.outputs}
        tensor_dict = {}
        for key in [
            "num_detections",
            "detection_boxes",
            "detection_scores",
            "detection_classes",
        ]:
            tensor_name = key + ":0"
            if tensor_name in all_tensor_names:
                tensor_dict[key] = tf.compat.v1.get_default_graph().get_tensor_by_name(
                    tensor_name
                )
        image_tensor = tf.compat.v1.get_default_graph().get_tensor_by_name("image_tensor:0")
        
        # Run inference
        output_dict_list = list()
        for frame in image_np_list:
            output_dict = sess.run(
                tensor_dict, feed_dict={image_tensor: np.expand_dims(frame, 0)}
            )
            # all outputs are float32 numpy arrays, so convert types as appropriate
            output_dict["num_detections"] = int(output_dict["num_detections"][0])
            output_dict["detection_classes"] = output_dict["detection_classes"][0].astype(np.uint8)
            output_dict["detection_boxes"] = output_dict["detection_boxes"][0]
            output_dict["detection_scores"] = output_dict["detection_scores"][0]
            output_dict_list.append(output_dict)

    return output_dict_list, image_np_list


@lru_cache()
def load_graph(PATH_TO_FROZEN_GRAPH):
    # ## Load a (frozen) Tensorflow model into memory.
    detection_graph = tf.Graph()
    with detection_graph.as_default():
        od_graph_def = tf.compat.v1.GraphDef()
        with tf.io.gfile.GFile(PATH_TO_FROZEN_GRAPH, 'rb') as fid:
            serialized_graph = fid.read()
            od_graph_def.ParseFromString(serialized_graph)
            tf.import_graph_def(od_graph_def, name='')
    return detection_graph


@lru_cache()
def videoframe(video)->list:
    frame_list = []
    test_video = cv2.VideoCapture(video)
    pos_frame = test_video.get(cv2.CAP_PROP_POS_FRAMES)
    while True:
        ret, frame = test_video.read()
        if (
            test_video.get(cv2.CAP_PROP_POS_FRAMES)
            == test_video.get(cv2.CAP_PROP_FRAME_COUNT)
            or test_video.get(cv2.CAP_PROP_FRAME_COUNT) == pos_frame
        ):
            break
        if ret:
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
            frame_list.append(frame)
            pos_frame = test_video.get(cv2.CAP_PROP_POS_FRAMES)
        else:
            pos_frame = (
                lambda x: test_video.get(cv2.CAP_PROP_FRAME_COUNT)
                if x >= test_video.get(cv2.CAP_PROP_FRAME_COUNT)
                else x
            )(pos_frame + 1)
            test_video.set(cv2.CAP_PROP_POS_FRAMES, pos_frame)
            continue

    return frame_list


class draw_bbox:
    def __init__(self, obj, draw=None, testing_mode=False):
        self.obj = obj
        self.obj = self.real_size()
        self.draw = draw
        self.color = self.get_color(obj["obj"])
        self.width = 9
        self.testing_mode = testing_mode

    def __call__(self):
        if self.draw:
            self.draw_box()
            self.draw_class_name()
            return True
        else:
            return False

    def draw_box(self):
        # 0: ymin, 1: xmin, 2: ymax, 3: xmax
        self.draw.rectangle(
            [
                self.obj["bbox"][1],self.obj["bbox"][0],
                self.obj["bbox"][3],self.obj["bbox"][2]
            ], 
            outline=self.color, 
            width=self.width)

    def draw_class_name(self):
        # ymin, xmin, ymax, xmax
        font_size = 30
        font = ImageFont.truetype(
            os.path.join(
                os.path.dirname(__file__),
                "font/HoonFuture_Bold.otf"), 
            font_size
        )
        if self.testing_mode:
            if self.obj['name'].split('_')[0] in 'grab pullout pick'.split():
                text = f"Action#{self.obj['obj']}: {100*self.obj['score']:.2f}%"
            else:
                text = f"Object#{self.obj['obj']}: {100*self.obj['score']:.2f}%"
        else:
            text = f"{self.obj['name']}: {100*self.obj['score']:.2f}%"
        text_size = font.getsize(text)

        self.draw.rectangle(
            [
                self.obj["bbox"][1],self.obj["bbox"][0],
                self.obj["bbox"][1]+text_size[0]+self.width,
                self.obj["bbox"][0]+text_size[1]
            ],
            fill=self.color, 
            )

        self.draw.text(
            (self.obj["bbox"][1]+self.width, self.obj["bbox"][0]),
            text,
            font=font,
            fill=(0, 0, 0, 255),
        )

    def rgbtobrg(self, rgb_color):
        r = (rgb_color % 16 ** 2) * 16 ** 4
        gb = rgb_color // 16 ** 2
        rgb = r + gb
        return rgb

    def get_color(self, obj_num):
        # RGB -> BRG
        obj_num = obj_num -1
        color_list = [
            0xF4E000,0x158E8E,0x6FF700,0xFF3232,0x000AD1,
            0xCBF400,0x15708E,0x6FF700,0xFF6532,0x2900D1,
            0x8EF400,0x15528E,0x6FF700,0xFF9932,0x5E00D1,
            0x40798C,0x70A9A1,0xF1AB86,0xF45B69,0xCA3CFF,
        ]
        return self.rgbtobrg(color_list[obj_num])

    def real_size(self):
        # ymin, xmin, ymax, xmax
        self.obj["bbox"][0] *= self.obj["size"][1]
        self.obj["bbox"][1] *= self.obj["size"][0]
        self.obj["bbox"][2] *= self.obj["size"][1]
        self.obj["bbox"][3] *= self.obj["size"][0]
        return self.obj


def class_int_to_text(class_num, label_map):
    class_list_num = class_num - 1
    camera = object_list(label_map)

    return camera.item_list()[class_list_num]


def object_to_json_style(cls, score, bbox, im, label_map):
    obj = dict()
    obj["obj"] = int(cls)
    obj["name"] = class_int_to_text(cls, label_map)
    obj["score"] = float("{:.4}".format(score))
    obj["bbox"] = [float(point) for point in bbox]
    obj["size"] = (im.shape[1],im.shape[0])

    return obj


class object_list:
    @lru_cache()
    def __init__(self, label_map):
        with open(label_map, 'r') as f:
            label_map = label_map_pb2.LabelMap()
            text_format.Merge(f.read(), label_map)
            self.objs = dict((i.name, 0) for i in label_map.item)

    def __repr__(self):
        return str(self.objs)

    def __setitem__(self, key, value):
        if key in self.objs.keys() and type(value) is int:
            self.objs[key] = value

    def __getitem__(self, item):
        return self.objs[item]

    def __len__(self):
        return len(self.objs)

    def make_feature(self):
        return [obj for obj in self.objs.values()]

    def item_list(self):
        return [key for key in self.objs.keys()]

    def item_group_list(self, category):
        return [obj_name for obj_name in self.item_list() if
                re.compile(f'({category})+$').search(obj_name)]